# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_core.ipynb.

# %% auto 0
__all__ = ['SLModel']

# %% ../nbs/02_core.ipynb 2
from .utils import *
from .imports import *
from .data import *

# %% ../nbs/02_core.ipynb 4
class SLModel(L.LightningModule):
    
    def __init__(self, model, learning_rate, cosine_t_max, in_chans=3, seed=42):
        super().__init__()
        L.seed_everything(seed)
        store_attr("model,learning_rate,cosine_t_max")
        self.example_input_array = torch.Tensor(1, in_chans, 224, 224)
        self.num_classes = model.num_classes
        self.save_hyperparameters(ignore=["model"])
        self.train_acc = TM.Accuracy(task="multiclass", num_classes=self.num_classes)
        self.val_acc = TM.Accuracy(task="multiclass", num_classes=self.num_classes)
        self.test_acc = TM.Accuracy(task="multiclass", num_classes=self.num_classes)

    def forward(self, x):
        return self.model(x)

    def _shared_step(self, batch):
        features, true_labels = batch['image'], batch['label']
        logits = self(features)
        loss = F.cross_entropy(logits, true_labels)
        predicted_labels = torch.argmax(logits, dim=1)
        return loss, true_labels, predicted_labels

    def training_step(self, batch, batch_idx):
        loss, true_labels, predicted_labels = self._shared_step(batch)
        self.log("train_loss", loss)
        self.train_acc(predicted_labels, true_labels)
        self.log(
            "train_acc", self.train_acc, prog_bar=True, on_epoch=True, on_step=False
        )
        return loss

    def validation_step(self, batch, batch_idx):
        loss, true_labels, predicted_labels = self._shared_step(batch)
        self.log("val_loss", loss, prog_bar=True)
        self.val_acc(predicted_labels, true_labels)
        self.log("val_acc", self.val_acc, prog_bar=True)

    def test_step(self, batch, batch_idx):
        loss, true_labels, predicted_labels = self._shared_step(batch)
        self.test_acc(predicted_labels, true_labels)
        self.log("test_acc", self.test_acc)

    def configure_optimizers(self):
        opt = optim.Adam(self.parameters(), lr=self.learning_rate)
        sch = torch.optim.lr_scheduler.CosineAnnealingLR(opt, T_max=self.cosine_t_max)
        return {"optimizer": opt, "lr_scheduler": {"scheduler": sch, "monitor": "train_loss",
                                                   "interval": "step", "frequency": 1}}



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/dreamai_dl',
                'doc_host': 'https://HamzaFarhan.github.io',
                'git_url': 'https://github.com/HamzaFarhan/dreamai_dl',
                'lib_path': 'dreamai_dl'},
  'syms': { 'dreamai_dl.core': { 'dreamai_dl.core.SLModel': ('core.html#slmodel', 'dreamai_dl/core.py'),
                                 'dreamai_dl.core.SLModel.__init__': ('core.html#slmodel.__init__', 'dreamai_dl/core.py'),
                                 'dreamai_dl.core.SLModel._shared_step': ('core.html#slmodel._shared_step', 'dreamai_dl/core.py'),
                                 'dreamai_dl.core.SLModel.configure_optimizers': ( 'core.html#slmodel.configure_optimizers',
                                                                                   'dreamai_dl/core.py'),
                                 'dreamai_dl.core.SLModel.forward': ('core.html#slmodel.forward', 'dreamai_dl/core.py'),
                                 'dreamai_dl.core.SLModel.test_step': ('core.html#slmodel.test_step', 'dreamai_dl/core.py'),
                                 'dreamai_dl.core.SLModel.training_step': ('core.html#slmodel.training_step', 'dreamai_dl/core.py'),
                                 'dreamai_dl.core.SLModel.validation_step': ('core.html#slmodel.validation_step', 'dreamai_dl/core.py')},
            'dreamai_dl.data': { 'dreamai_dl.data.SLDataModule': ('data.html#sldatamodule', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataModule.__init__': ('data.html#sldatamodule.__init__', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataModule.prepare_data': ('data.html#sldatamodule.prepare_data', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataModule.setup': ('data.html#sldatamodule.setup', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataModule.test_dataloader': ( 'data.html#sldatamodule.test_dataloader',
                                                                                   'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataModule.train_dataloader': ( 'data.html#sldatamodule.train_dataloader',
                                                                                    'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataModule.val_dataloader': ( 'data.html#sldatamodule.val_dataloader',
                                                                                  'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataset': ('data.html#sldataset', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataset.__getitem__': ('data.html#sldataset.__getitem__', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataset.__init__': ('data.html#sldataset.__init__', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataset.__len__': ('data.html#sldataset.__len__', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataset.get_at_index': ('data.html#sldataset.get_at_index', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataset.get_img': ('data.html#sldataset.get_img', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataset.get_img_name': ('data.html#sldataset.get_img_name', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataset.get_img_path': ('data.html#sldataset.get_img_path', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataset.get_label': ('data.html#sldataset.get_label', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.SLDataset.get_tensor': ('data.html#sldataset.get_tensor', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.StatsDataset': ('data.html#statsdataset', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.StatsDataset.__getitem__': ('data.html#statsdataset.__getitem__', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.StatsDataset.__init__': ('data.html#statsdataset.__init__', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.StatsDataset.__len__': ('data.html#statsdataset.__len__', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.get_data_stats': ('data.html#get_data_stats', 'dreamai_dl/data.py'),
                                 'dreamai_dl.data.split_df': ('data.html#split_df', 'dreamai_dl/data.py')},
            'dreamai_dl.imports': {},
            'dreamai_dl.utils': { 'dreamai_dl.utils.UnNormalize': ('utils.html#unnormalize', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.UnNormalize.__init__': ('utils.html#unnormalize.__init__', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.del_norm': ('utils.html#del_norm', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.freeze_model': ('utils.html#freeze_model', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.freeze_params': ('utils.html#freeze_params', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.get_norm': ('utils.html#get_norm', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.get_norm_id': ('utils.html#get_norm_id', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.is_frozen': ('utils.html#is_frozen', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.is_norm': ('utils.html#is_norm', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.is_sequential': ('utils.html#is_sequential', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.is_tensor': ('utils.html#is_tensor', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.is_unfrozen': ('utils.html#is_unfrozen', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.params': ('utils.html#params', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.plot_loss_and_acc': ('utils.html#plot_loss_and_acc', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.plt_show': ('utils.html#plt_show', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.show_img': ('utils.html#show_img', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.tensor_to_img': ('utils.html#tensor_to_img', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.unfreeze_model': ('utils.html#unfreeze_model', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.unfreeze_params': ('utils.html#unfreeze_params', 'dreamai_dl/utils.py'),
                                  'dreamai_dl.utils.update_norm': ('utils.html#update_norm', 'dreamai_dl/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_utils.ipynb.

# %% auto 0
__all__ = ['UnNormalize', 'is_sequential', 'params', 'is_frozen', 'is_unfrozen', 'freeze_params', 'unfreeze_params',
           'freeze_model', 'unfreeze_model', 'get_norm', 'is_norm', 'get_norm_id', 'del_norm', 'update_norm',
           'is_tensor', 'tensor_to_img', 'plt_show', 'show_img', 'plot_loss_and_acc']

# %% ../nbs/00_utils.ipynb 2
from .imports import *

# %% ../nbs/00_utils.ipynb 4
class UnNormalize(transforms.Normalize):
    def __init__(self,mean,std,*args,**kwargs):
        mean = [-m/s for m,s in zip(mean,std)]
        std = [1/s for s in std]
        super().__init__(mean, std, *args, **kwargs)

def is_sequential(x):
    return isinstance(x, nn.Sequential)

def params(m):
    return [p for p in m.parameters()]

def is_frozen(model):
    return np.array([not p.requires_grad for p in (params(model))]).all()

def is_unfrozen(model):
    return np.array([p.requires_grad for p in (params(model))]).all()

def freeze_params(params):
    for p in params:
        p.requires_grad = False

def unfreeze_params(params):
    for p in params:
        p.requires_grad = True

def freeze_model(model):
    freeze_params(model.parameters())

def unfreeze_model(model):
    unfreeze_params(model.parameters())

def get_norm(tfms):
    if tfms is None:
        return tfms
    try:
        tfms_list = list(tfms)
    except:
        tfms_list = list(tfms.transforms)
    for t in tfms_list:
        if is_norm(t):
            return t
    return None

def is_norm(x):
    return type(x).__name__ == 'Normalize'

def get_norm_id(tfms):
    try:
        tfms_list = list(tfms)
    except:
        tfms_list = tfms.transforms
    for i,t in enumerate(tfms_list):
        if is_norm(t):
            return t,i
    return None, None

def del_norm(tfms, idx=None):
    if idx is None:
        idx = get_norm_id(tfms)[1]
    if idx is not None:
        del tfms.transforms[idx]

def update_norm(tfms, mean, std, idx=None):
    if idx is None:
        idx = get_norm_id(tfms)[1]
    if idx is not None:
        tfms.transforms[idx].mean = mean
        tfms.transforms[idx].std = std
    else:
        tfms.transforms.append(transforms.Normalize(mean=mean, std=std))

def is_tensor(x):
    return isinstance(x, torch.Tensor)

def tensor_to_img(t):
    if t.dim() > 3:
        return [np.array(np.transpose(t_,(1,2,0))) for t_ in t]
    return np.array(np.transpose(t,(1,2,0)))

def plt_show(im, cmap=None, title='', figsize=(7,7)):
    if path_or_str(im):
        im = rgb_read(im)
    if is_tensor(im):
        im = tensor_to_img(im)
        if is_list(im): im = im[0]
    fig=plt.figure(figsize=figsize)
    plt.imshow(im, cmap=cmap)
    plt.title(title)
    plt.show()

def show_img(img, cmap=None, titles=[''], figsize=(7,7)):
    if not is_list(img):
        imgs = [img]
    else:
        imgs = img
    [plt_show(img, cmap=cmap, title=title, figsize=figsize) for img,title in zip(imgs,titles)]

def plot_loss_and_acc(log_dir, loss_ylim=(0.0, 0.9), acc_ylim=(0.7, 1.0), save_loss=None, save_acc=None):

    metrics = pd.read_csv(f"{log_dir}/metrics.csv")

    aggreg_metrics = []
    agg_col = "epoch"
    for i, dfg in metrics.groupby(agg_col):
        agg = dict(dfg.mean())
        agg[agg_col] = i
        aggreg_metrics.append(agg)

    df_metrics = pd.DataFrame(aggreg_metrics)
    df_metrics[["train_loss", "val_loss"]].plot(
        grid=True, legend=True, xlabel="Epoch", ylabel="Loss"
    )

    plt.ylim(loss_ylim)
    if save_loss is not None:
        plt.savefig(save_loss)

    df_metrics[["train_acc", "val_acc"]].plot(
        grid=True, legend=True, xlabel="Epoch", ylabel="ACC"
    )

    plt.ylim(acc_ylim)
    if save_acc is not None:
        plt.savefig(save_acc)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_data.ipynb.

# %% auto 0
__all__ = ['split_df', 'get_data_stats', 'StatsDataset', 'SLDataset', 'SLDataModule']

# %% ../nbs/01_data.ipynb 2
from .imports import *
from .utils import *

# %% ../nbs/01_data.ipynb 4
def split_df(train_df, test_size=0.15, stratify_idx=1):
    try:    
        train_df,test_df = train_test_split(train_df, test_size=test_size, stratify=train_df.iloc[:,stratify_idx])
    except:
        print('Not stratified.')
        train_df,test_df = train_test_split(train_df, test_size=test_size)
    train_df = train_df.reset_index(drop=True)
    test_df = test_df.reset_index(drop=True)
    return train_df,test_df

def get_data_stats(df, data_path='', img_idx=0, img_size=224, channels=3, stats_percentage=0.7, bs=32, num_workers=4, device=None):
    
    if device is None:
        device = default_device()
    print('Calculating dataset mean and std. This may take a while.\n')
    frac_data = df.copy().sample(frac=stats_percentage).reset_index(drop=True).copy()
    if list_or_tuple(img_size):
        img_size = img_size[0]
    tfms = transforms.Compose([transforms.Resize((img_size,img_size)), transforms.ToTensor()])
    dset = StatsDataset(frac_data, data_path=data_path, tfms=tfms, img_idx=img_idx, channels=channels)
    dl = DataLoader(dset, batch_size=bs, num_workers=num_workers)
    batches = len(dl)
    mean = 0.0
    print('Mean loop:')
    for i,data_batch in enumerate(dl):
        try:
            if i % (batches//10) == 0:
                print(f'Batch: {i+1}/{batches}')
        except:
            print(f'Batch: {i+1}/{batches}')
        # images = data_batch[0]
        if is_list(data_batch['image']):
            images = data_batch['image'][0].to(device)
        else:
            images = data_batch['image'].to(device)
        batch_samples = images.size(0) 
        images = images.view(batch_samples, images.size(1), -1)
        mean += images.to(dtype=torch.float32).mean(2).sum(0)
    mean = mean / len(dl.dataset)
    var = 0.0
    print('\nStd loop:')
    for i,data_batch in enumerate(dl):
        try:
            if i % (batches//10) == 0:
                print(f'Batch: {i+1}/{batches}')
        except:
            print(f'Batch: {i+1}/{batches}')
        # images = data_batch[0]
        if is_list(data_batch['image']):
            images = data_batch['image'][0].to(device)
        else:
            images = data_batch['image'].to(device)
        batch_samples = images.size(0)
        images = images.view(batch_samples, images.size(1), -1)
        var += ((images - mean.unsqueeze(1))**2).sum([0,2])
    std = torch.sqrt(var / (len(dl.dataset)*img_size*img_size))
    print('\nDone.')
    return mean.cpu(), std.cpu()

class StatsDataset(Dataset):
    def __init__(self, data, data_path='', img_idx=0, tfms=None, channels=3, **kwargs):
        super().__init__()
        if tfms is None:
            tfms = transforms.Compose([transforms.ToTensor()])
        data_path = Path(data_path)
        store_attr(','.join(dict_keys(locals_to_params(locals()))))
    
    def __len__(self):
        return len(self.data)
    
    def __getitem__(self, index):
        img = Image.open(self.data_path/self.data.iloc[index, self.img_idx])
        if self.channels == 1:
            img = img.convert('L')
        else:
            img = img.convert('RGB')
        return {'image': self.tfms(img)}

class SLDataset(Dataset):
    def __init__(self, data, data_path='', tfms=None, img_idx=0, label_idx=1, channels=3, class_names=None, **kwargs):
        super().__init__()
        
        data_path = Path(data_path)
        if class_names is None:
            class_names = np.unique(data.iloc[:, label_idx])
        idx_to_class = dict(enumerate(class_names))
        class_to_idx = {v:k for k,v in idx_to_class.items()}
             
        store_attr(','.join(dict_keys(locals_to_params(locals()))))
        
    def __len__(self):
        return len(self.data)
    
    def get_img_name(self, index):
        # return [self.data.iloc[index, x] for x in self.img_idx]
        return self.data.iloc[index, self.img_idx]
    
    def get_img_path(self, index):
        return str(self.data_path/self.get_img_name(index))
        # return [str(self.data_path/x) for x in self.get_img_names(index)]
        
    def get_img(self, index):
        img = Image.open(self.get_img_path(index))
        if self.channels == 1:
            return img.convert('L')
        return img.convert('RGB')
        # return [Image.open(x).convert('RGB') for x in self.get_img_paths(index)]
    
    def get_tensor(self, index, to_tensor=True):
        img = self.get_img(index)
        if self.tfms is not None:
            x = self.tfms(copy.deepcopy(img))
            # if self.channels == 1:
                # x.unsqueeze_(0)
            if not to_tensor:
                x = tensor_to_img(x)
        else:
            x = img
        return x
    
    def get_label(self, index, get_idx=True):
        if get_idx:
            return self.class_to_idx[self.data.iloc[index, self.label_idx]]
            # return [cid[self.data.iloc[index, x]] for cid,x in zip(self.class_to_idx,self.label_idx)]
        return self.data.iloc[index, self.label_idx]
        # return [self.data.iloc[index, x] for x in self.label_idx]

    def __getitem__(self, index, to_tensor=True, get_idx=True, get_name=False, get_path=False):
        ret = {'image':self.get_tensor(index, to_tensor=to_tensor), 'label':self.get_label(index, get_idx=get_idx)}
        if get_name:
            ret['image_name'] = self.get_img_name(index)
        if get_path:
            ret['image_path'] = self.get_img_path(index)
        return ret

    # def denorm_data(self, data, keys=['images']):
    #     if self.tfms is not None:
    #         norm_t = get_norm(self.tfms)
    #         if norm_t:
    #             mean = norm_t.mean
    #             std = norm_t.std
    #             for k in keys:
    #                 data[k] = denorm_img(data[k], mean, std)

    def get_at_index(self, index, to_tensor=False, get_name=True, get_path=True):

        data = self.__getitem__(index, to_tensor=to_tensor, get_idx=False, get_name=get_name, get_path=get_path)
        # if denorm:
            # self.denorm_data(data=data)
        return data
    
class SLDataModule(L.LightningDataModule):
    def __init__(self, parq_file, data_path='', train_tfms=None, test_tfms=None, train_size=0.8, valid_size=0.2, test_size=0.1,
                 img_size=512, channels=3, batch_size=64, num_workers=6, pin_memory=False, calc_stats=False, stats_img_size=224,
                 stats_file='img_stats.pkl', seed=42, **kwargs):
        super().__init__()
        L.seed_everything(seed)
        self.save_hyperparameters(ignore=['parq_file', 'data_path'])
        if is_list(img_size):
                img_size = tuple(img_size)
        test_size = test_size/valid_size
        stats_file = Path(data_path)/stats_file
        store_attr(','.join(dict_keys(locals_to_params(locals()))))

    def prepare_data(self):
        if self.calc_stats:
            df = pd.read_parquet(self.parq_file)
            self.img_mean, self.img_std = get_data_stats(df, data_path=self.data_path, img_size=self.stats_img_size,
                                                         stats_percentage=0.7, bs=64, channels=self.channels,
                                                         num_workers=self.num_workers, device='cpu')
            save_obj(self.stats_file, [self.img_mean, self.img_std])
        elif Path(self.stats_file).exists() and get_norm(self.train_tfms) is None:
            self.img_mean, self.img_std = load_obj(self.stats_file)
        elif get_norm(self.train_tfms) is not None:
            norm = get_norm(self.train_tfms)
            self.img_mean, self.img_std = norm.mean, norm.std
        else:
            self.img_mean, self.img_std = imagenet_stats
        if self.train_tfms is None:
            self.train_tfms = transforms.Compose([transforms.Resize(self.img_size),
                                                  transforms.RandomHorizontalFlip(), transforms.RandomVerticalFlip(),
                                                  transforms.ToTensor()])
        if self.test_tfms is None:
            self.test_tfms = transforms.Compose([transforms.Resize(self.img_size), transforms.ToTensor()])
        
        update_norm(self.train_tfms, self.img_mean, self.img_std)
        update_norm(self.test_tfms, self.img_mean, self.img_std)
    
    def setup(self, stage=None):
        df = pd.read_parquet(self.parq_file)
        train_df, valid_df = split_df(df, test_size=self.valid_size)
        valid_df, test_df = split_df(valid_df, test_size=self.test_size)
        self.train_ds = SLDataset(train_df, data_path=self.data_path, tfms=self.train_tfms, channels=self.channels)
        self.valid_ds = SLDataset(valid_df, data_path=self.data_path, tfms=self.test_tfms, channels=self.channels)
        self.test_ds = SLDataset(test_df, data_path=self.data_path, tfms=self.test_tfms, channels=self.channels)
        self.class_names = self.train_ds.class_names
        self.num_classes = len(self.class_names)
        self.idx_to_class = self.train_ds.idx_to_class

    def train_dataloader(self):
        return DataLoader(dataset=self.train_ds, batch_size=self.batch_size, drop_last=True,
                          shuffle=True, num_workers=self.num_workers, pin_memory=self.pin_memory)
    def val_dataloader(self):
        return DataLoader(dataset=self.valid_ds, batch_size=self.batch_size, drop_last=False,
                          shuffle=False, num_workers=self.num_workers, pin_memory=self.pin_memory)

    def test_dataloader(self):
        return DataLoader(dataset=self.test_ds, batch_size=self.batch_size, drop_last=False,
                          shuffle=False, num_workers=self.num_workers, pin_memory=self.pin_memory)



import timm
import torch
import torch.nn as nn
import lightning as L
from torch import optim
import torchmetrics as TM
import torch.nn.functional as F
from torchvision.utils import make_grid
import lightning.pytorch.callbacks as cb
from lightning.pytorch.tuner import Tuner
from torchvision import transforms, models
import torchvision.transforms.functional as TF
from lightning.pytorch.loggers import CSVLogger
from torch.utils.data import Dataset, DataLoader
from torch.utils.data.dataset import random_split
from sklearn.model_selection import train_test_split
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.utilities.model_summary import ModelSummary


from dreamai.core import *
from dreamai.vision import *
from dreamai.imports import *

from fastai.vision.core import imagenet_stats
from fastai.vision.core import image2tensor as img_to_tensor
from fastai.torch_core import Module, default_device, to_device
from fastai.vision.learner import has_pool_type, create_head, cut_model
from fastai.layers import LinBnDrop, Flatten, AdaptiveConcatPool2d, SigmoidRange